@@ -1,6 +1,6 @@
 /*
  * Normaliz
- * Copyright (C) 2007-2014  Winfried Bruns, Bogdan Ichim, Christof Soeger
+ * Copyright (C) 2007-2019  Winfried Bruns, Bogdan Ichim, Christof Soeger
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
@@ -21,17 +21,20 @@
  * terms of service.
  */
 
-#ifndef CONE_H_
-#define CONE_H_
+#ifndef LIBNORMALIZ_CONE_H_
+#define LIBNORMALIZ_CONE_H_
 
 #include <sys/stat.h>
 #include <vector>
 #include <map>
+#include <set>
+#include <string>
 #include <utility> //for pair
 #include <boost/dynamic_bitset.hpp>
 
-#include <libnormaliz/libnormaliz.h>
-#include <libnormaliz/cone_property.h>
+#include <libnormaliz/general.h>
+#include "libnormaliz/input_type.h"
+#include <libnormaliz/automorph.h>
 #include <libnormaliz/sublattice_representation.h>
 #include <libnormaliz/matrix.h>
 #include <libnormaliz/HilbertSeries.h>
@@ -42,14 +45,56 @@ using std::map;
 using std::pair;
 
 template<typename Integer> class Full_Cone;
-//template<typename Integer> class Matrix;
+
+
+template<typename Integer>
+struct FACETDATA{
+    
+        
+        vector<Integer> Hyp;               // linear form of the hyperplane
+        boost::dynamic_bitset<> GenInHyp;  // incidence hyperplane/generators
+        Integer ValNewGen;                 // value of linear form on the generator to be added
+        size_t BornAt;                      // number of generator (in order of insertion) at which this hyperplane was added,, counting from 0
+        size_t Ident;                      // unique number identifying the hyperplane (derived from HypCounter)
+        size_t Mother;                     // Ident of positive mother if known, 0 if unknown
+        bool is_positive_on_all_original_gens;
+        bool is_negative_on_some_original_gen;
+        bool simplicial;                   // indicates whether facet is simplicial
+    };
+    
+template<typename Integer>
+struct CONVEXHULLDATA{
+    
+    Sublattice_Representation<Integer> SLR; // identifies the version of BasisChangePointed with which the data were stored
+    long nr_threads;
+    
+    bool is_primal;
+
+    vector<size_t> HypCounter; // counters used to give unique number to hyperplane
+                               // must be defined thread wise to avoid critical
+                               
+    vector<bool> in_triang;  // intriang[i]==true means that Generators[i] has been actively inserted
+    vector<key_t> GensInCone;    // lists the generators completely built in
+    size_t nrGensInCone;    // their number
+    
+    vector<size_t> Comparisons; // at index i we note the total number of comparisons 
+                               // of positive and negative hyperplanes needed for the first i generators
+    size_t nrTotalComparisons; // counts the comparisons in the current computation
+        
+
+    list<FACETDATA<Integer> > Facets;  // contains the data for Fourier-Motzkin and extension of triangulation
+    size_t old_nr_supp_hyps; // must be remembered since Facets gets extended before the current generators is finished
+    Matrix<Integer> Generators;
+};
 
 // type for simplex, short in contrast to class Simplex
-template<typename Integer> struct SHORTSIMPLEX {
+template<typename Integer> 
+struct SHORTSIMPLEX {
     vector<key_t> key;                // full key of simplex
     Integer height;                   // height of last vertex over opposite facet
     Integer vol;                      // volume if computed, 0 else
-    vector<bool> Excluded;           // for disjoint decomposition of cone
+    Integer vol_for_detsum;           // used by QuFullCone
+    vector<bool> Excluded;            // for disjoint decomposition of cone
                                       // true in position i indictate sthat the facet 
                                       // opposite of generator i must be excluded
 };
@@ -84,6 +129,8 @@ public:
 //                    Constructors, they preprocess the input
 //---------------------------------------------------------------------------
     
+    typedef Integer elem_type;
+
     Cone(); //default constructor
 
     /* give up to 3 matrices as input
@@ -174,8 +221,7 @@ public:
          InputType type3, const Matrix<nmz_float>& input_data3);
 
     /* give multiple input */
-    Cone(const map< InputType , Matrix<nmz_float> >& multi_input_data);
-    
+    Cone(const map< InputType , Matrix<nmz_float> >& multi_input_data);    
     
 //---------------------------------------------------------------------------
 //                                Destructor
@@ -187,6 +233,7 @@ public:
 //                          give additional data
 //---------------------------------------------------------------------------
 
+
     /* Sets if the Cone prints verbose output.
      * The default value for the Cone is the global verbose.
      * returns the old value
@@ -194,7 +241,54 @@ public:
     bool setVerbose (bool v);
 
     void deactivateChangeOfPrecision();
+    
+    /* We allow the change of the cone by additional inequalities or generators
+     * after the first computation for "dynamical" applications, in which
+     * thecone is canged depending on previous computation results.
+     * 
+     * Only one additional matrix can be added at a time. Nevertheless it is useful
+     * to go through the map. 
+     */
 
+    void modifyCone(const map< InputType , vector<vector<Integer> > >& add_multi_input);
+    void modifyCone(const map< InputType , vector<vector<mpq_class> > >& add_multi_input);
+    void modifyCone(const map< InputType , vector<vector<nmz_float> > >& add_multi_input);
+    
+    void modifyCone(InputType type, const vector< vector<Integer> >& input_data);
+    void modifyCone(InputType type, const vector< vector<mpq_class> >& input_data);
+    void modifyCone(InputType type, const vector< vector<nmz_float> >& input_data);
+    
+    void modifyCone(InputType type, const Matrix<Integer>& input_data);
+    void modifyCone(InputType type, const Matrix<mpq_class>& input_data);
+    void modifyCone(InputType type, const Matrix<nmz_float>& input_data);
+    
+    /* We must also transport data that cannot be conveyed by the constructors
+     * or comute functions (in the present setting)
+     */
+    
+    void set_project(string name);
+    void set_nmz_call(const string& path);
+    void set_output_dir(string name);
+    
+    void setPolynomial(string poly);
+    
+    void setNumericalParams(const map <NumParam::Param, long >& num_params);
+    void setNrCoeffQuasiPol(long nr_coeff);
+    void setExpansionDegree(long degree);
+    void setFaceCodimBound(long bound);
+    void setAutomCodimBoundMult(long bound);
+    void setAutomCodimBoundVectors(long bound);
+
+    void setRenf(renf_class *renf);
+
+    template<typename InputNumber>    
+    void check_add_input(const map< InputType, vector< vector<InputNumber> > >& multi_add_data);
+    template<typename InputNumber>
+    void check_consistency_of_dimension(const map< InputType, vector< vector<InputNumber> > >& multi_add_data);
+
+    map< InputType, vector< vector<Integer> > > mpqclass_input_to_integer(
+          const map< InputType, vector< vector<mpq_class> > >& multi_input_data_const);
+    
 //---------------------------------------------------------------------------
 //                           make computations
 //---------------------------------------------------------------------------
@@ -292,6 +386,16 @@ public:
     const Matrix<Integer>& getDeg1ElementsMatrix();
     const vector< vector<Integer> >& getDeg1Elements();
     size_t getNrDeg1Elements();
+    
+    size_t getNumberLatticePoints();
+    
+    const Matrix<Integer>& getLatticePointsMatrix();
+    const vector< vector<Integer> >& getLatticePoints();
+    size_t getNrLatticePoints();
+    
+    const map<boost::dynamic_bitset<>,int>& getFaceLattice();
+    vector<size_t> getFVector();
+    const vector<boost::dynamic_bitset<> >& getIncidence();
 
     // the actual grading is Grading/GradingDenom
     vector<Integer> getGrading();
@@ -300,12 +404,18 @@ public:
     vector<Integer> getDehomogenization();
 
     vector<Integer> getClassGroup();
+    
+    const AutomorphismGroup<Integer>& getAutomorphismGroup(ConeProperty::Enum quality);
+    const AutomorphismGroup<Integer>& getAutomorphismGroup();
 
     mpq_class getMultiplicity();
     mpq_class getVolume();
+    renf_elem_class getRenfVolume();
     nmz_float getEuclideanVolume();
     mpq_class getVirtualMultiplicity();
     mpq_class getIntegral();
+    nmz_float getEuclideanIntegral();
+
     const pair<HilbertSeries, mpz_class>& getWeightedEhrhartSeries();
     
     string getPolynomial() const;
@@ -326,6 +436,7 @@ public:
 
     const Sublattice_Representation<Integer>& getSublattice();
     const HilbertSeries& getHilbertSeries(); //general purpose object
+    const HilbertSeries& getEhrhartSeries(); //general purpose object
 
     // the following 2 methods give information about the last used triangulation
     // if no triangulation was computed so far they return false
@@ -338,31 +449,58 @@ public:
     list< STANLEYDATA_int >& getStanleyDec_mutable(); //allows us to erase the StanleyDec
                              // in order to save memeory for weighted Ehrhart
     
-    void set_project(string name);
-    void set_nmz_call(const string& path);
-    void set_output_dir(string name);
-    
-    void setPolynomial(string poly);
-    void setNrCoeffQuasiPol(long nr_coeff);
-    void setExpansionDegree(long degree);
-    
     bool get_verbose ();
+    void write_cone_output(const string& output_file);
     
     IntegrationData& getIntData();
 
+    void resetGrading(vector<Integer> lf);
+
+    const Matrix<Integer>& getMatrixConePropertyMatrix(ConeProperty::Enum property);
+    const vector< vector<Integer> >& getMatrixConeProperty(ConeProperty::Enum property);
+
+    const Matrix<nmz_float>& getFloatMatrixConePropertyMatrix(ConeProperty::Enum property);
+    const vector< vector<nmz_float> >& getFloatMatrixConeProperty(ConeProperty::Enum property);
+    
+    vector<Integer> getVectorConeProperty(ConeProperty::Enum property);
+
+    Integer getIntegerConeProperty(ConeProperty::Enum property);
+
+    mpz_class getGMPIntegerConeProperty(ConeProperty::Enum property);
+    
+    mpq_class getRationalConeProperty(ConeProperty::Enum property);
+
+    nmz_float getFloatConeProperty(ConeProperty::Enum property);
+    
+    renf_elem_class getFieldElemConeProperty(ConeProperty::Enum property);
+
+    size_t getMachineIntegerConeProperty(ConeProperty::Enum property);
+
+    bool getBooleanConeProperty(ConeProperty::Enum property);
+    
+    nmz_float euclidean_corr_factor();
+
 //---------------------------------------------------------------------------
 //                          private part
 //---------------------------------------------------------------------------
 
+    
 private:
     
     string project;
     string output_dir;
     string nmz_call;
     size_t dim;
+    bool inhom_input;
     
-    // the following three matrices store the constraints of the input
+    bool keep_convex_hull_data; // indicates that data computed in Full_Cone and other data are preserved and can be used again
+    CONVEXHULLDATA<Integer> ConvHullData;
+    bool conversion_done; // indicates that generators have been converted to inequalities
+    
+    // the following matrices store the constraints of the input
     Matrix<Integer> Inequalities;
+    Matrix<Integer> AddInequalities; // for inequalities added later on
+    Matrix<Integer> AddGenerators; // for inequalities added later on
     Matrix<Integer> Equations;
     Matrix<Integer> Congruences;
     // we must register some information about thew input
@@ -377,7 +515,9 @@ private:
     // Matrix<Integer> GeneratorsOfToricRing;
     Matrix<Integer> OriginalMonoidGenerators;
     Matrix<Integer> Generators;
-    Matrix<Integer> ExtremeRays;
+    Matrix<Integer> ReferenceGenerators;
+    Matrix<Integer> ExtremeRays; // of the homogenized cone
+    Matrix<Integer> ExtremeRaysRecCone; // of the recession cone, = ExtremeRays in the homogeneous case
     Matrix<nmz_float> VerticesFloat;
     vector<bool> ExtremeRaysIndicator;
     Matrix<Integer> VerticesOfPolyhedron;
@@ -398,36 +538,51 @@ private:
     mpq_class multiplicity;
     mpq_class volume;
     nmz_float euclidean_volume;
+    nmz_float euclidean_height; // for volume computations wuth renf_elem_class
+    renf_elem_class renf_volume;
     mpq_class Integral;
     mpq_class VirtualMultiplicity;
     vector<Integer> WitnessNotIntegrallyClosed;
     vector<Integer> GeneratorOfInterior;
     Matrix<Integer> HilbertBasis;
-    // Matrix<Integer> HilbertBasisRecCone;
+    Matrix<Integer> HilbertBasisRecCone;
     Matrix<Integer> BasisMaxSubspace;
     Matrix<Integer> ModuleGeneratorsOverOriginalMonoid;
     Matrix<Integer> Deg1Elements;
     HilbertSeries HSeries;
+    HilbertSeries EhrSeries;
     IntegrationData IntData;
     vector<Integer> Grading;
     vector<Integer> Dehomogenization;
+    vector<Integer> Norm; // used by v_standardize in the numberfield case
     Integer GradingDenom;
     Integer index;  // the internal index
     Integer unit_group_index;
+    size_t number_lattice_points;
+    vector<size_t> f_vector;
     
     vector<boost::dynamic_bitset<> > Pair; // for indicator vectors in project-and_lift
     vector<boost::dynamic_bitset<> > ParaInPair; // if polytope is a parallelotope
     bool check_parallelotope();
     bool is_parallelotope;
+    
+    map<boost::dynamic_bitset<>,int> FaceLattice;
+    vector<boost::dynamic_bitset<> > SuppHypInd; // incidemnce vectors of the support hyperplanes
 
     bool pointed;
     bool inhomogeneous;
+
+    bool input_automorphisms;
+
+    bool polytope_in_input;
     bool gorensetin;
+
     bool deg1_extreme_rays;
     bool deg1_hilbert_basis;
     bool integrally_closed;
     bool Gorenstein;
     bool rees_primary;
+    bool dual_original_generators; // true means: dual cone has original generators
     Integer ReesPrimaryMultiplicity;
     int affine_dim; //dimension of polyhedron
     size_t recession_rank; // rank of recession monoid
@@ -441,12 +596,22 @@ private:
     Matrix<Integer> WeightsGrad;
     vector<bool> GradAbs;
 
-    bool no_lattice_restriction; // true if cine generators are known to be in the relevant lattice
     bool normalization; // true if input type normalization is used
+    bool general_no_grading_denom;
+    
+#ifdef ENFNORMALIZ    
+    renf_class *Renf;
+#endif
+    
+    long renf_degree;
+    long face_codim_bound;
 
     // if this is true we allow to change to a smaller integer type in the computation
     bool change_integer_type;
     
+    long autom_codim_vectors;
+    long autom_codim_mult;
+
     Cone<Integer>* IntHullCone; // cone containing data of integer hull
     Cone<Integer>* SymmCone;    // cone containing symmetrized data
     Cone<Integer>* ProjCone;    // cone containing projection to selected coordinates 
@@ -467,9 +632,13 @@ private:
     void prepare_input_constraints(const map< InputType, vector< vector<Integer> > >& multi_input_data);
     void prepare_input_generators(map< InputType, vector< vector<Integer> > >& multi_input_data,
                      Matrix<Integer>& LatticeGenerators);
-    void homogenize_input(map< InputType, vector< vector<Integer> > >& multi_input_data);
+    template<typename InputNumber>
+    void homogenize_input(map< InputType, vector< vector<InputNumber> > >& multi_input_data);
     void check_precomputed_support_hyperplanes();
     void check_excluded_faces();
+    bool check_lattice_restrictions_on_generators(bool& cone_sat_cong);
+    
+    void check_gens_vs_reference(); //to make sure that newly computed generators agrre with the previously computed
     
     void setGrading (const vector<Integer>& lf);
     void setWeights ();
@@ -481,6 +650,15 @@ private:
     
     void try_symmetrization(ConeProperties& ToCompute);
     void try_approximation_or_projection (ConeProperties& ToCompute);
+    
+    void try_Hilbert_Series_from_lattice_points(const ConeProperties& ToCompute);
+    void make_Hilbert_series_from_pos_and_neg(const vector<num_t>& h_vec_pos, const vector<num_t>& h_vec_neg);
+    
+    void make_face_lattice(const ConeProperties& ToCompute);
+    void compute_combinatorial_automorphisms(const ConeProperties& ToCompute);
+    void compute_euclidean_automorphisms(const ConeProperties& ToCompute);
+    
+    AutomorphismGroup<Integer> Automs;
 
     Matrix<Integer> prepare_input_type_2(const vector< vector<Integer> >& Input);
     Matrix<Integer> prepare_input_type_3(const vector< vector<Integer> >& Input);
@@ -493,9 +671,9 @@ private:
     void compute_full_cone(ConeProperties& ToCompute);
 
     /* compute the generators using the support hyperplanes */
-    void compute_generators();
+    void compute_generators(ConeProperties& ToCompute);
     template<typename IntegerFC>
-    void compute_generators_inner();
+    void compute_generators_inner(ConeProperties& ToCompute);
 
     /* compute method for the dual_mode, used in compute(mode) */
     void compute_dual(ConeProperties& ToCompute);
@@ -506,17 +684,23 @@ private:
 
     /* extract the data from Full_Cone, this may remove data from Full_Cone!*/
     template<typename IntegerFC>
-    void extract_data(Full_Cone<IntegerFC>& FC);
-    template<typename IntegerFC>
-    void extract_supphyps(Full_Cone<IntegerFC>& FC);
+    void extract_data(Full_Cone<IntegerFC>& FC,ConeProperties& ToCompute);
     
-    void extract_supphyps(Full_Cone<Integer>& FC);
-
+    template<typename IntegerFC>
+    void extract_convex_hull_data(Full_Cone<IntegerFC>& FC, bool primal); 
+    template<typename IntegerFC>
+    void push_convex_hull_data(Full_Cone<IntegerFC>& FC, bool primal);
+    
+    template<typename IntegerFC>
+    void extract_supphyps(Full_Cone<IntegerFC>& FC, Matrix<Integer>& ret, bool dual=true);
+    void extract_supphyps(Full_Cone<Integer>& FC, Matrix<Integer>& ret, bool dual=true);    
+    
+    void norm_dehomogenization(size_t FC_dim);
 
     /* set OriginalMonoidGenerators */
     void set_original_monoid_generators(const Matrix<Integer>&);
 
-    /* set ExtremeRays, in inhomogeneous case also VerticesOfPolyhedron */
+    /* set ExtremeRays, in inhomogeneous case also VerticesOfPolyhedron and ExtremeRaysRecCone*/
     void set_extreme_rays(const vector<bool>&);
 
     /* If the Hilbert basis and the original monoid generators are computed,
@@ -536,6 +720,8 @@ private:
     void complete_sublattice_comp(ConeProperties& ToCompute); // completes the sublattice computations
     void complete_HilbertSeries_comp(ConeProperties& ToCompute);
     
+    void treat_polytope_as_being_hom_defined(ConeProperties ToCompute);
+    
     void compute_integral (ConeProperties& ToCompute);
     void compute_virt_mult (ConeProperties& ToCompute);
     void compute_weighted_Ehrhart(ConeProperties& ToCompute);
@@ -549,23 +735,38 @@ private:
     
     void set_parallelization();
     
+    void handle_dynamic(const ConeProperties& ToCompute);
+    
     template<typename IntegerFC>
     void give_data_of_approximated_cone_to(Full_Cone<IntegerFC>& FC);
     
-    void project_and_lift(ConeProperties& ToCompute, Matrix<Integer>& Deg1, const Matrix<Integer>& Gens, Matrix<Integer>& Supps, bool float_projection);
+    void project_and_lift(const ConeProperties& ToCompute, Matrix<Integer>& Deg1, const Matrix<Integer>& Gens, 
+                          const Matrix<Integer>& Supps, const Matrix<Integer>& Congs, const vector<Integer> GradingOnPolytope); 
 
     void compute_volume(ConeProperties& ToCompute);
-    void compute_euclidean_volume(const vector<Integer>& Grad, Integer GradDenom);
     
     void try_multiplicity_by_descent(ConeProperties& ToCompute);
+    void try_multiplicity_of_para(ConeProperties& ToCompute);
     
     void compute_projection(ConeProperties& ToCompute);
     void compute_projection_from_gens(const vector<Integer>& GradOrDehom);
-    void compute_projection_from_constraints(const vector<Integer>& GradOrDehom);
+    void compute_projection_from_constraints(const vector<Integer>& GradOrDehom, ConeProperties& ToCompute);
  
     //in order to avoid getRank fromm inside compute
     size_t get_rank_internal();
     const Sublattice_Representation<Integer>& get_sublattice_internal();
+    
+    void compute_lattice_points_in_polytope(ConeProperties& ToCompute);
+    void prepare_volume_computation(ConeProperties& ToCompute);
+    
+    bool set_quality_of_automorphisms(ConeProperties& ToCompute, AutomParam::Quality& quality_of_automorphisms);
+    
+    template<typename IntegerFC>
+    vector<vector<key_t> > extract_permutations(const vector<vector<key_t> >&  FC_Permutations,
+                Matrix<IntegerFC>& FC_Vectors, const Matrix<Integer>& ConeVectors, bool primal, vector<key_t>& Key);
+
+    vector<vector<key_t> > extract_subsets(const vector<vector<key_t> >& FC_Subsets, size_t max_index,const vector<key_t>& Key);
+
 };
 
 // helpers
@@ -583,7 +784,7 @@ void insert_column(vector< vector<Integer> >& mat, size_t col, Integer entry);
 // computes approximating lattice simplex using the A_n dissection of the unit cube
 // q is a rational vector with the denominator in the FIRST component q[0]
 template<typename Integer>
-void approx_simplex(const vector<Integer>& q, std::list<vector<Integer> >& approx, const long approx_level){
+inline void approx_simplex(const vector<Integer>& q, std::list<vector<Integer> >& approx, const long approx_level){
 	
 	//cout << "approximate the point " << q;
     long dim=q.size();
@@ -621,7 +822,7 @@ void approx_simplex(const vector<Integer>& q, std::list<vector<Integer> >& appro
 	}
 	//cout << "the best level is " << (best_level+1) << endl;
 	//now we proceed as before
-	vector<pair<Integer,size_t>> best_remain(dim);
+	vector<pair<Integer,size_t> > best_remain(dim);
 	for(long i=0;i<dim;i++){
 		best_remain[i].first = remain[best_level][i];
 		best_remain[i].second = i; // after sorting we must lnow where elements come from
@@ -651,6 +852,10 @@ void approx_simplex(const vector<Integer>& q, std::list<vector<Integer> >& appro
 
 }
 
+template<>
+inline void approx_simplex(const vector<renf_elem_class>& q, std::list<vector<renf_elem_class> >& approx, const long approx_level){
+    assert(false);    
+}
 
 }  //end namespace libnormaliz
 
